/***
 * Excerpted from "Modern Systems Programming with Scala Native",
 * published by The Pragmatic Bookshelf.
 * Copyrights apply to this code. It may not be used to create training material,
 * courses, books, articles, and the like. Contact us if you are in doubt.
 * We make no guarantees that this code is fit for any purpose.
 * Visit http://www.pragmaticprogrammer.com/titles/rwscala for more book information.
***/
import scala.scalanative.unsafe._
import scala.scalanative.libc._

// import scalanative.native._
import scala.scalanative.posix.unistd

case class Command(path:String, args:String, env:Map[String,String])
object main {
  import util._
  def main(args:Array[String]):Unit = {
    if (args.size == 0) {
      println("bye")
      stdlib.exit(1)
    }

    println("about to fork")

    val status = doAndAwait { () =>
      println(s"in child, about to exec command: ${args.toSeq}")
      runCommand(args)
    }
    println(s"wait status ${status}")
  }

  def doFork(task:Function0[Int]):Int = {
    val pid = fork()
    if (pid > 0) {
      pid
    } else {
      val res = task.apply()
      stdlib.exit(res)
      res
    }
  }

  def await(pid:Int):Int = {
    val status = stackalloc[Int]
    waitpid(pid, status, 0)
    val statusCode = !status
    if (statusCode != 0) {
      throw new Exception(s"Child process returned error $statusCode")
    }
    !status
  }

  def doAndAwait(task:Function0[Int]):Int = {
    val pid = doFork(task)
    await(pid)
  }

  def runCommand(args:Seq[String], env:Map[String,String] = Map.empty):Int = {
    if (args.size == 0) {
      throw new Exception("bad arguments of length 0")
    }
    Zone { implicit z =>
      val fname = toCString(args.head)
      val arg_array = makeStringArray(args)
      val env_strings = env.map { case (k,v) => s"$k=$v" }
      val env_array = makeStringArray(env_strings.toSeq)

      val r = execve(fname, arg_array, env_array)
      if (r != 0) {
        val err = errno.errno
        stdio.printf(c"error: %d %d\n", err, string.strerror(err) )
        throw new Exception(s"bad execve: returned $r")
      }
    }
    ??? // This will never be reached.
  }

  def makeStringArray(args:Seq[String]):Ptr[CString] = {
    val pid = unistd.getpid()
    val size = sizeof[Ptr[CString]] * args.size + 1
    val dest_array = stdlib.malloc(size).asInstanceOf[Ptr[CString]]
    val count = args.size
    Zone { implicit z =>
      for ( (arg,i) <- args.zipWithIndex) {
        val string_ptr = toCString(arg)
        val string_len = string.strlen(string_ptr)
        val dest_str = stdlib.malloc(string_len + 1).asInstanceOf[Ptr[Byte]]
        string.strncpy(dest_str, string_ptr, arg.size + 1)
        dest_str(string_len) = 0
        dest_array(i) = dest_str
        ()
      }
      ()
    }
    dest_array(count) = null
    for (j <- (0 to count)) {
    }
    dest_array
  }

def runOneAtATime(commands:Seq[Seq[String]]) = {
  for (command <- commands) {
    doAndAwait { () =>
      runCommand(command)
    }
  }
}

def runSimultaneously(commands:Seq[Seq[String]]) = {
  val pids = for (command <- commands) yield {
    doFork { () =>
      runCommand(command)
    }
  }
  for (pid <- pids) {
    await(pid)
  }
}

def awaitAny(pids:Set[Int]):Set[Int] = {
  val status = stackalloc[Int]
  var running = pids
  !status = 0
  val finished = waitpid(-1, status, 0)
  if (running.contains(finished)) {
    val statusCode = !status
    if (statusCode != 0) {
      throw new Exception(s"Child process returned error $statusCode")
    } else {
      return pids - finished
    }
  } else {
    throw new Exception(s"""error: reaped process ${finished},
                        expected one of $pids""")
  }
}

def awaitAll(pids:Set[Int]):Unit = {
  var running = pids
  while (running.nonEmpty) {
    println(s"waiting for $running")
    running = awaitAny(running)
  }
  println("Done!")
}


// TODO
def badThrottle(commands:Seq[Seq[String]], maxParallel:Int) = {
  ???
}

// TODO
def goodThrottle(commands:Seq[Seq[String]], maxParallel:Int) = {
  ???
}
}

@extern
object util {
  def execve(filename:CString, args:Ptr[CString], env:Ptr[CString]):Int = extern
  def execvp(path:CString, args:Ptr[CString]):Int = extern
  def fork():Int = extern
  def getpid():Int = extern
  def waitpid(pid:Int, status:Ptr[Int], options:Int):Int = extern
  def strerror(errno:Int):CString = extern

  def pipe(pipes:Ptr[Int]):Int = extern
  def dup2(oldfd:Int, newfd:Int):Int = extern
}
